#! /usr/bin/env python
"""Set a random background image (XWin)

Select a random image from a (list of) directory(s)
and set it as the desktop wallpaper (display it in the root window)
using xli program.

"""

__version__ = "$Revision$"[11:-2]
__revision__ = "$Id$"[5:-2]
__date__ = "$Date$"[7:-2]

__author__ = "Oleg Broytman <phd@phdru.name>"
__copyright__ = "Copyright (C) 2000-2010 PhiloSoft Design"
__license__ = "GNU GPL"


import sys, os
from ConfigParser import SafeConfigParser

import anydbm
import random
import shelve
import subprocess
from time import time
from fcntl import flock, LOCK_EX, LOCK_UN, LOCK_NB

from wsgiref import simple_server
from wsgiref.handlers import SimpleHandler
from wsgiref.simple_server import WSGIServer, make_server
simple_server.ServerHandler = SimpleHandler # Stop logging to stdout


def error(error_str, error_code=1):
   sys.stderr.write("%s: Error: %s\n" % (sys.argv[0], error_str))
   sys.exit(error_code)


xsetbg_dir = os.path.dirname(os.path.abspath(__file__))
os.chdir(xsetbg_dir)

config = SafeConfigParser()
config.read("xsetbg.conf")

if config.has_option("images", "directory") or \
   config.has_option("images", "directory0") or \
   config.has_option("images", "directory1"):
   image_dirs = []
   if config.has_option("images", "directory"):
      image_dirs.append(config.get("images", "directory"))
   if config.has_option("images", "directory0"):
      image_dirs.append(config.get("images", "directory0"))
   if config.has_option("images", "directory1"):
      image_dirs.append(config.get("images", "directory1"))
   i = 2
   while True:
      option = "directory%d" % i
      if config.has_option("images", option):
         image_dirs.append(config.get("images", option))
         i += 1
      else:
         break
else:
   image_dirs = ["images"]

image_dirs = [os.path.join(xsetbg_dir,
   os.path.expandvars(os.path.expanduser(dirname)))
      for dirname in image_dirs
]

# minimum time in seconds between background image changes
if config.has_option("xsetbg", "min_pause"):
   min_pause = config.getint("xsetbg", "min_pause")
else:
   min_pause = 60

borders = config.get("xsetbg", "borders").split(',')
if config.has_option("xsetbg", "borders"):
   borders = [border.strip() for border in config.get("xsetbg", "borders").split(',')]
else:
   borders = ["darkcyan", "steelblue", "midnightblue"]

# minimum time in seconds between occurences of the same image
if config.has_option("xsetbg", "min_delay"):
   min_delay = config.getint("xsetbg", "min_delay")
else:
   min_delay = 3600*24 # 24 hours

# httpd settings
if config.has_option("httpd", "host"):
   host = config.get("httpd", "host")
else:
   host = 'localhost'

if config.has_option("httpd", "port"):
   port = config.getint("httpd", "port")
else:
   error("Config must specify a port to listen. Abort.")

del config


os.umask(0066) # octal; -rw-------; make the global persistent dictionary
               # readable only by the user
global_db_name = "xsetbg.db"


# DB keys
timestamp_key = "timestamp"
filename_key = "filename"
old_filename_key = "old_filename"


# Create the database if it is not exists yet.

try:
   global_db = shelve.open(global_db_name, flag='c')
except anydbm.error, msg:
   if str(msg) == "db type could not be determined":
      os.remove(global_db_name)
      global_db = shelve.open(global_db_name, flag='c')

# Remove old filenames
old_time = time() - min_delay

to_delete = [timestamp_key]
for key in global_db.keys():
   if key.startswith('/') and global_db[key] < old_time:
      to_delete.append(key)

for key in to_delete:
   try:
      del global_db[key]
   except KeyError:
      pass

global_db.close() # Close DB in the parent process


images = []

for image_dir in image_dirs:
   # List images in all subdirectories
   for dirpath, dirs, files in os.walk(image_dir):
      images.extend([os.path.join(dirpath, file) for file in files])

if not images:
   error("No images found. Abort.")


def published(func):
    func._wsgi_published = True
    return func

@published
def ping(force=False):
   # Use the program's file as the lock file:
   # lock it to prevent two processes run in parallel.
   lock_file = open("xsetbg.py", 'r')
   try:
      flock(lock_file, LOCK_EX|LOCK_NB)
   except IOError: # already locked
      lock_file.close()
      return


   global_db = None
   try:
      # Reopen the global persistent dictionary
      global_db = shelve.open(global_db_name, 'w')

      timestamp = global_db.get(timestamp_key)
      current_time = time()

      if not force and timestamp is not None and \
         current_time - timestamp < min_pause: # Too early to change background
            return

      # Save current time
      global_db[timestamp_key] = current_time

      # Select a random image and check if we've seen it recently;
      # loop until we can find a new image (never seen before) or old enough.
      for i in xrange(len(images)): # ensure the loop is not infinite
         image_name = random.choice(images)
         if global_db.has_key(image_name):
            image_time = global_db[image_name]
            if current_time - image_time > min_delay:
               break
         else:
            break
      global_db[image_name] = current_time

      # Save filename
      if global_db.has_key(filename_key):
         global_db[old_filename_key] = global_db[filename_key]
      global_db[filename_key] = image_name

      program_options = ["xli", "-onroot", "-quiet"] + \
         ["-center", "-border", random.choice(borders), "-zoom", "auto",
            image_name]

      rc = subprocess.call(program_options)
      if rc:
         error("cannot execute xli!")

   finally:
      # Unlock and close the lock file
      flock(lock_file, LOCK_UN)
      lock_file.close()
      # Flush and close the global persistent dictionary
      if global_db: global_db.close()

@published
def force():
    ping(force=True)

@published
def stop():
    QuitWSGIServer._quit_flag = True


g = globals().copy()
commands = dict([(name, g[name]) for name in g
    if getattr(g[name], '_wsgi_published', False)])
del g

class QuitWSGIServer(WSGIServer):
    _quit_flag = False

    def serve_forever(self):
        while not self._quit_flag:
            self.handle_request()

def app(env, start_response):
    command = env['PATH_INFO'][1:] # Remove the leading slash
    if command not in commands:
        status = '404 Not found'
        response_headers = [('Content-type', 'text/plain')]
        start_response(status, response_headers)
        return ['The command was not found.\n']

    try:
        commands[command]()
    except:
        status = '500 Error'
        response_headers = [('Content-type', 'text/plain')]
        start_response(status, response_headers)
        return ['Error occured!\n']

    status = '200 OK'
    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)
    return ['Ok\n']

force()
httpd = make_server(host, port, app, server_class=QuitWSGIServer)
httpd.serve_forever()
